<!DOCTYPE html>
<html lang="en">
<head>
    <meta charset="utf-8">
    <title>Oxigraph server</title>
    <link href="./yasgui.min.css" rel="stylesheet" type="text/css" />
    <script src="./yasgui.min.js"></script>
    <link rel="icon" type="image/svg+xml" href="/logo.svg">
</head>
<body>
<div id="yasgui"></div>
<script>
    const url = window.location.href.endsWith('/') ? window.location.href.slice(0, -1) : window.location.href;
    new Yasgui(document.getElementById("yasgui"), {
        requestConfig: { endpoint: url + "/sparql" },
        endpointCatalogueOptions: {
            getData: function () {
                return [
                    { endpoint: url + "/sparql" },
                    { endpoint: url + "/query" },
                    { endpoint: url + "/update" },
                ];
            },
            keys: [],
        }
    });
</script>
</body>
